"""Main implementation"""
from abc import ABC
import dimod
import random
from typing import Any, Dict, List, Tuple, TypeVar

T = TypeVar("T")


def random_bit_value():
    """Returns random bit"""
    return random.randint(0, 1)


class StructuredSampler(dimod.Sampler, dimod.Structured, ABC):
    pass


class SamplerMock(dimod.Sampler):
    """Class mocking arbitrary dimod unstructured sampler."""

    def __init__(
        self,
        properties: Dict[Any, Any],
        parameters: Dict[Any, List[Any]],
        get_random_bit=random_bit_value,
    ):
        self._properties = properties
        self._parameters = parameters
        self.get_random_bit = get_random_bit

    def sample(self, bqm: dimod.BinaryQuadraticModel, **parameters) -> dimod.SampleSet:
        for parameter in parameters:
            if parameter not in self.parameters:
                raise TypeError(f"Parameter {parameter} is not supported by this sampler.")

        if bqm.vartype == dimod.Vartype.SPIN:
            get_random_value = lambda: self.get_random_bit() * 2 - 1
        else:
            get_random_value = self.get_random_bit

        samples = [
            {variable: get_random_value() for variable in bqm.variables}
            for _ in range(parameters.get("num_reads", 1))
        ]

        return dimod.SampleSet.from_samples_bqm(samples, bqm)

    @property
    def parameters(self) -> Dict[Any, List[Any]]:
        """Returns dict of parameters of this mock-sampler.

        See dimod's documentation for further reference on this property.
        """
        return self._parameters

    @property
    def properties(self) -> Dict[Any, Any]:
        """Returns dict of properties of this sampler.

        See dimod's documentation for further reference on this parameter.
        """
        return self._properties

    @classmethod
    def from_sampler(cls, sampler: dimod.Sampler) -> "SamplerMock":
        return cls(properties=sampler.properties, parameters=sampler.parameters)


class StructuredMock(SamplerMock, dimod.Structured):
    """Class mocking arbitrary-structured sampler."""

    def __init__(
        self,
        nodelist: List[T],
        edgelist: List[Tuple[T, T]],
        properties: Dict[Any, Any],
        parameters: Dict[Any, List[Any]],
    ):
        super().__init__(properties, parameters)
        self._nodelist = nodelist
        self._edgelist = edgelist

    @dimod.decorators.bqm_structured
    def sample(self, bqm, **parameters) -> dimod.SampleSet:
        return super().sample(bqm, **parameters)

    @property
    def nodelist(self) -> List[T]:
        """Returns list of all nodes of this mock sampler."""
        return self._nodelist

    @property
    def edgelist(self) -> List[Tuple[T, T]]:
        """Returns list of all edges of this mock sampler."""
        return self._edgelist

    @classmethod
    def from_sampler(cls, sampler: StructuredSampler) -> "SamplerMock":
        return cls(
            nodelist=sampler.nodelist,
            edgelist=sampler.edgelist,
            properties=sampler.properties,
            parameters=sampler.parameters,
        )



""""Test cases for StructuredMock"""
import pytest
from dimodmock import StructuredMock

EXAMPLE_CONFIGURATIONS = [
    ([0, 1, 2, 3], [(0, 2), (1, 3)], {}, {}),
    (list(range(1000)), [(0, 1), (1, 2), (0, 100)], {}, {}),
    (
        list(range(1000)),
        [(0, 1000), (10, 20)],
        {"max_num_reads": 100},
        {"num_reads": ["max_num_reads"]},
    ),
    (
        ["a", "b", "c", "d"],
        [("a", "b"), ("c", "a")],
        {"max_num_reads": 100},
        {"num_reads": ["max_num_reads"]},
    ),
]


@pytest.mark.parametrize("nodelist,edgelist,properties,parameters", EXAMPLE_CONFIGURATIONS)
def test_propagates_nodelist(nodelist, edgelist, properties, parameters):
    mock = StructuredMock(nodelist, edgelist, properties, parameters)
    assert mock.nodelist == nodelist


@pytest.mark.parametrize("nodelist,edgelist,properties,parameters", EXAMPLE_CONFIGURATIONS)
def test_propagates_edgelist(nodelist, edgelist, properties, parameters):
    mock = StructuredMock(nodelist, edgelist, properties, parameters)
    assert mock.edgelist == edgelist


@pytest.mark.parametrize("nodelist,edgelist,properties,parameters", EXAMPLE_CONFIGURATIONS)
def test_propagates_properties(nodelist, edgelist, properties, parameters):
    mock = StructuredMock(nodelist, edgelist, properties, parameters)
    assert mock.properties == properties


@pytest.mark.parametrize("nodelist,edgelist,properties,parameters", EXAMPLE_CONFIGURATIONS)
def test_propagates_parameters(nodelist, edgelist, properties, parameters):
    mock = StructuredMock(nodelist, edgelist, properties, parameters)
    assert mock.parameters == parameters



"""Test cases for SamplerMock"""
import pytest
import dimod
from dimodmock import SamplerMock


EXAMPLE_CONFIGURATIONS = [
    ({}, {}),
    ({"max_num_reads": 100}, {"num_reads": ["max_num_reads"], "iterations": []}),
    ({"j_range": [-1, 1]}, {"num_reads": []}),
]


@pytest.mark.parametrize("properties,parameters", EXAMPLE_CONFIGURATIONS)
def test_propagates_parameters(properties, parameters):
    """SamplerMock should correctly return dict of parameters specified during initialization."""
    mock = SamplerMock(properties, parameters)
    assert mock.parameters == parameters


@pytest.mark.parametrize("properties,parameters", EXAMPLE_CONFIGURATIONS)
def test_propagates_properties(properties, parameters):
    """SamplerMock should correctly return dict of properties specified during initialization."""
    mock = SamplerMock(properties, parameters)
    assert mock.properties == properties


def test_rejects_unknown_parameters():
    """SamplerMock should raise TypeError if unknown parameters are passed to sample* methods."""
    mock = SamplerMock({}, {"num_reads": []})

    with pytest.raises(TypeError):
        mock.sample_qubo({(0, 1): 10}, num_reads=10, annealing_time=1000)


def test_accepts_known_properties():
    """"SamplerMock should not raise if only known parameters are passed for sample* methods."""
    mock = SamplerMock({}, {"num_reads": [], "annealing_time": []})

    try:
        mock.sample_qubo({(0, 0): 5}, num_reads=5, annealing_time=10)
    except TypeError:
        pytest.fail("SamplerMock unexpectedly raised TypeError.")


@pytest.mark.parametrize(
    "vartype,random_bits,expected_variables_values",
    [
        (dimod.Vartype.SPIN, [0, 1, 0], [-1, 1, -1]),
        (dimod.Vartype.BINARY, [0, 1, 1, 1, 1, 0], [0, 1, 1, 1, 1, 0]),
    ],
)
def tests_uses_get_random_bit(mocker, vartype, random_bits, expected_variables_values):
    get_random_bit = mocker.Mock(side_effect=random_bits)
    mock = SamplerMock({}, {}, get_random_bit=get_random_bit)
    bqm = dimod.BinaryQuadraticModel(
        linear={i: i for i in range(len(random_bits))}, quadratic={}, offset=0, vartype=vartype
    )
    result = mock.sample(bqm)
    assert all(
        result.first.sample[i] == expected_variables_values[i] for i in range(len(random_bits))
    )


@pytest.mark.parametrize("num_reads", [1, 10, 20, 100])
def test_respects_num_reads(num_reads):
    """SampleMock.sample should correctly use num_reads parameter."""
    mock = SamplerMock({}, {"num_reads": []})
    result = mock.sample_qubo({(0, 1): 1, (0, 0): 2}, num_reads=num_reads)
    assert len(result) == num_reads



"""Setup file for dimodmock package"""
from setuptools import setup, find_packages

with open("README.md") as description_file:
    long_description = description_file.read()


setup(
    name="dimodmock",
    use_scm_version=True,
    description="Utilities for mocking dimod Samplers",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Konrad Jałowiecki",
    author_email="dexter2206@gmail.com",
    packages=find_packages(exclude="tests"),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=["dimod"],
    tests_require=["dimod", "pytest", "pytest-mock"],
    setup_requires=["setuptools_scm"],
    keywords="dimod mock testing"
)
